# -*- coding: utf-8 -*-

"""Class that parses the Wireguard configuration files as Python dictionaries"""

import collections
import logging
import os
import pprint


logger = logging.getLogger(__name__);


class WireguardConfig():
    '''Class for parsing the WireGuard config files'''

    def __init__(self):
        '''Constructor'''
        self.configpath = '/etc/wireguard'
        self.clear_data()

    def clear_data(self):
        '''Clears all data so that it can be set anew'''
        self.data = collections.defaultdict(dict)
        
    def parse_wg_config(self, interface, prefix=''):
        '''Parses the given WireGuard config and stores it'''
        
        def section2dict(section_lines, prefix):
            '''Converts the lines of a section into a dictionary'''
            result = dict()
            for line in section_lines:
                k,_, v = line.partition('=')
                k = k.strip().lower()
                v = v.strip()
                if k == 'allowedips':
                    k = 'allowed-ips'
                    v = v.replace(' ', '').split(',')
                elif k == 'persistentkeepalive':
                    k = 'persistent-keepalive'
                elif k == 'preshared-key':
                    k = 'preshared-key'
                elif k == 'privatekey':
                    k = 'private-key'
                elif k == 'publickey':
                    k = 'public-key'
                v = int(v) if k in ['persistent-keepalive'] else v # convert strings to integers for certain attributes
                result[prefix + k] = v
            return result
        
        # Note: ConfigParser cannot be used since it does not support duplicate "[Peer]" sections; thus do it ourselves
        with open(os.path.join(self.configpath, interface+'.conf'), 'r') as file:
            config = file.readlines()
        config = [ line.strip() for line in config ] # strip newlines, etc.
        config = [ line for line in config if (len(line)>0) and (line[0] not in [';', '#']) ] # strip comments, empty lines, etc.
        config = [ line if line[0] != '[' else line.lower() for line in config ] # section names to lower case
        config.append('[eof]')
        section_isinterface = None
        section_lines = []
        for line in config:
            if line[0] == '[':
                if section_isinterface is not None: # nothing to do on start of first section
                    d = section2dict(section_lines, prefix)
                    if section_isinterface:
                        self.data[interface] = {**d, **self.data[interface]} # merge dictionaries
                        self.data[interface]['peers'] = self.data[interface].get('peers', dict())
                    else:
                        k = d.get(prefix + 'public-key', '[PublicKey missing]')
                        self.data[interface]['peers'][k] = {**d, **self.data[interface]['peers'].get(k, dict())} # merge dictionaries
                # Prepare for reading section further
                section_isinterface = (line == '[interface]')
                section_lines = []
            else:
                section_lines.append(line)

    def retrieve_wireguard_config(self, interfaces, data=None, prefix=''):
        '''Sets the local data based on parsed WireGuard config files'''
        if data is None:
            self.clear_data()
        else:
            self.data = data
        for interface in interfaces:
            try:
                self.parse_wg_config(interface, prefix)
            except FileNotFoundError: # ignore missing config files (interface could have been configured by other means)
                logger.info('No config file found for interface [{0}]'.format(interface))

    @property
    def wgconfig():
        return self.data

    def get_interfaces(self):
        '''Returns the list of interfaces (as an iterator)'''
        return self.data.keys()
        
    def get_interfacedata(self, interface):
        '''Returns the data of the given interface as dictionary'''
        return self.data.get(interface)
        
    def get_peerdata(self, interface):
        '''Returns the data of the peers of the given interface as dictionary'''
        return self.data.get(interface, dict()).get('peers')


if __name__ == '__main__':
    wg = WireguardConfig()
    wg.retrieve_wireguard_config(['wg_1'])
    pprint.pprint(wg.get_interfaces())
    pprint.pprint(wg.get_interfacedata('wg_1'))
    pprint.pprint(wg.get_peerdata('wg_1'))



# -*- coding: utf-8 -*-

import asyncio
import logging
import socket

from . import datakeeper as dk
from . import output


logger = logging.getLogger(__name__)


class Logic():
    '''Class that contains the business logic of this application'''

    def __init__(self, config, func_enqueue):
        '''Constructor'''
        self.config = config
        self.func_enqueue = func_enqueue
        self.data = dk.DataKeeper(config)

    def initialize_data(self):
        '''Reload the config and status'''
        self.data.initialize()

    async def ping(self, destination, ping6=False):
        '''Asynchronously execute the ping command to chech reachability'''
        command = 'ping'
        if ping6 or (':' in destination):
            command = 'ping6'
        proc = await asyncio.create_subprocess_exec(command, '-q', '-c', '1', '-w', '1', destination, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
        stdout, stderr = await proc.communicate()
        return (proc.returncode != 0)

    def is_hostname(self, peername):
        '''Checks whether the provided peer is defined by hostname (in contrast to IP address)'''
        if peername is None:
            return False
        result = not (':' in peername) # ":" indicates an IPv6 address and is not allowed in hostname
        if result: # still a hostname candidate?
            parts = peername.split('.')
            result = (len(parts) != 4) or not (parts[0].isnumeric()) or not (parts[1].isnumeric()) or not (parts[2].isnumeric()) or not (parts[3].isnumeric())
        return result

    def endpoint_is_hostname(self, endpoint):
        '''Checks whether the provided endpoint is defined by hostname (in contrast to IP address)'''
        if endpoint is None:
            return False
        return self.is_hostname(endpoint.partition(':')[0])

    async def do_periodically(self):
        '''Tasks to be executed periodically each cycle (called by scheduler coroutine)'''
        logger.debug('Executing periodic tasks')
        # Updates the WireGuard status information
        self.data.update_status()
        # Get config attributes
        cycles_wait = self.config.cycles_wait
        cycles_checking = self.config.cycles_checking
        cycles_checkperiod = self.config.cycles_checkperiod
        cycles_slowcheckingperiod = self.config.cycles_slowcheckingperiod
        ping_interval = self.config.ping_interval
        ping_failafternum = self.config.ping_failafternum
        # Iterate through all peers of all interfaces and determine new status
        ping_plan = []
        for interface, interfacedata, peer, peerdata in self.data.peeriterator():
            status = peerdata.get('status', 'undefined')
            cycle_counter = peerdata.get('cycle-counter', 0)
            #print(interface, 'Status', status, cycle_counter)
            update_peer = False
            # Act based on current state
            next = 'unchanged'
            if peerdata.get('handshake-status', 'failed') != 'failed':
                if peerdata.get('ping-address') is None:
                    peerdata['ping-address'] = peerdata['allowed-ips'][0].partition(':')[0]
                if ping_interval > 0:
                    if cycle_counter % ping_interval == 0:
                        next = 'ping'                
                if (status == 'undefined') and (next == 'unchanged'):
                    next = 'up:ok'
            elif (status == 'undefined') or (status == 'up:ok'):
                if self.endpoint_is_hostname(peerdata.get('config_endpoint')):
                    next = 'down:waiting' if (cycles_wait > 0) else 'down:checking'
                else: # no further check of peer needed
                    if peerdata.get('endpoint') is None:
                        next = 'disabled'
                    else:
                        next = 'down'
            elif status == 'down:waiting':
                if cycle_counter >= cycles_wait:
                    next = 'down:checking'
            elif status == 'down:checking':
                if cycle_counter >= cycles_checking:
                    next = 'down:backingoff'
                else:
                    if cycle_counter % cycles_checkperiod == 0:
                        update_peer = True
            elif status == 'down:backingoff':
                if cycle_counter >= peerdata['backingoff-limit']:
                    update_peer = True
                    peerdata['cycle-counter'] = 0
                    peerdata['backingoff-limit'] = 2 * peerdata['backingoff-limit']
                    if peerdata['backingoff-limit'] >= cycles_slowcheckingperiod:
                        next = 'down:slowchecking'
            elif status == 'down:slowchecking':
                if cycle_counter >= cycles_slowcheckingperiod:
                    update_peer = True
                    next = 'down:slowchecking'
            elif status == 'down':               
                pass
            elif status == 'disabled':               
                pass
            else:
                logger.critical('Unknown status [{0}]'.format(status))
            # Count cycles
            peerdata['cycle-counter'] = peerdata.get('cycle-counter', 0) + 1
            # Change state as needed
            if next == 'ping':
                if status == 'undefined':
                    peerdata['cycle-counter'] = 0
                    status = 'down:checking'
                ping_plan.append((interface, interfacedata, peer, peerdata))
                next = None
            elif next == 'down:backingoff':
                peerdata['backingoff-limit'] = 2 * cycles_checkperiod
            elif next == 'down:slowchecking':
                peerdata['backingoff-limit'] = None
            elif next == 'unchanged':
                next = None
            if next is not None:
                logger.info('Changing status of [{interface}:{peer}] to [{next}] after {cycle_counter} cycles'.format(interface=interface, peer=peer, next=next, cycle_counter=cycle_counter))
                peerdata['cycle-counter'] = 0
                status = next
            # Update peer if requested
            if update_peer:
                config_endpoint = peerdata.get('config_endpoint')
                logger.debug('Requesting to check for update of [{interface}:{peer}], endpoint [{config_endpoint}]'.format(interface=interface, peer=peer, config_endpoint=config_endpoint))
                await self.func_enqueue('update_peer', { 'interface': interface, 'peer': peer, 'config_endpoint': peerdata.get('config_endpoint'), 'endpoint': peerdata.get('endpoint') })
            #self.data.set(interface, peer, 'status', status)
            peerdata['status'] = status
        # Check reachability by pinging peers
        if len(ping_plan) > 0:
            ping_tasks = []
            for interface, interfacedata, peer, peerdata in ping_plan:
                addr = peerdata['ping-address']
                ping_tasks.append(asyncio.ensure_future(self.ping(addr)))
            await asyncio.gather(*ping_tasks)
            for i, ping_task in enumerate(ping_tasks):
                if ping_task.result():
                    if ping_plan[i][3]['status'] != 'up:ok':
                        ping_plan[i][3]['status'] = 'up:ok'
                        ping_plan[i][3]['cycle-counter'] = 0
                        logger.info('Changing status of [{interface}:{peer}] to [up:ok] after successful ping'.format(interface=ping_plan[i][0], peer=ping_plan[i][2]))
                    ping_plan[i][3]['ping-failcounter'] = 0
                else:
                    ping_plan[i][3]['ping-failcounter'] = ping_plan[i][3].get('ping-failcounter', 0) + 1
                    if ping_plan[i][3]['ping-failcounter'] >= 'ping_failafternum':
                        logger.info('Changing status of [{interface}:{peer}] to [down:waiting] after failed ping'.format(interface=ping_plan[i][0], peer=ping_plan[i][2]))
                        ping_plan[i][3]['status'] = 'down:waiting'
                        ping_plan[i][3]['cycle-counter'] = 0
        # Output new status
        await output.output_status(self.config.outputs, self.data)

    def update_peer(self, interface, peer, config_endpoint, endpoint):
        '''Checks whether peer needs to be updated and does it if needed'''
        config_endpoint, _, config_port = config_endpoint.partition(':')
        endpoint = endpoint.partition(':')[0]
        logger.info('Resolving [{0}]'.format(config_endpoint))
        needed_endpoint = socket.getaddrinfo(config_endpoint, 0)[0][4][0] # get ip address
        if needed_endpoint != endpoint:
            logger.info('Changing interface endpoint [{0}] to changed IP [{1}]'.format(config_endpoint, needed_endpoint))        
            if config_port is not None:
                needed_endpoint += ':' + config_port
            self.data.set_endpoint(interface, peer, needed_endpoint)

    async def process_queue(self, item):
        '''Process an item from the event queue (called by queue listener coroutine)'''
        data = item.get('data', dict())
        if item.get('command') == 'update_peer':
            self.update_peer(data['interface'], data['peer'], data['config_endpoint'], data['endpoint'])
        else:
            logger.critical('Unknown command in event [{0}]'.format(item.get('command')))



# -*- coding: utf-8 -*-

"""Class that uses the 'wg show' command to get the status of WireGuard as Python dictionaries"""


import collections
import logging
import pprint
import shlex
import subprocess
import time


logger = logging.getLogger(__name__);


class WireguardCommand():
    '''Class for parsing the output of WireGuard's dump command'''

    def __init__(self, interface='all'):
        '''Constructor'''
        self.interface = interface
        self.clear_data()

    def clear_data(self):
        '''Clears all data so that it can be set anew'''
        self.data = collections.defaultdict(dict)

    def check_handshake(self, latest_handshake, persistent_keepalive):
        '''Checks whether handshake succeeded as expected'''
        if latest_handshake == 0: # no handshake took place successfully yet
            return None, False
        if persistent_keepalive == None: # convert this parameter to an integer
            persistent_keepalive = 0
        delta = int(time.time()) - latest_handshake # delta in seconds
        # Based on a forum: 2 minutes + keepalive + 2s (error margin)
        # However, this does not seem to match https://www.wireguard.com/papers/wireguard.pdf
        if delta <= 120:
            handshake_status = 'ok'
        elif delta <= 135: # 135 is used by the WireGuard reresolve-dns.sh script
            handshake_status = 'pending'
        elif delta <= 180:
            handshake_status = 'retrying'
        else:
            handshake_status = 'failed'
        return delta, handshake_status

    def execute(self, command, suppressoutput=False, suppresserrors=False):
        '''Execute a command'''
        args = shlex.split(command);
        nsp = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE);
        out, err = nsp.communicate();
        if err is not None:
            err = err.decode('utf8');
            if not suppresserrors and (len(err) > 0):
                logger.error(err);
        out = out.decode('utf8');
        if not suppressoutput and (len(out) > 0):
            print(out);       
        nsp.wait();
        return out, err;

    def execute_wg_set(self, interface, peer, attr, value, suppressoutput=True, suppresserrors=False):
        '''Execute the WireGuard command to set the provided attribute'''
        command = 'wg set "{0}"'.format(interface)
        if peer is not None:
            command += ' peer "{0}"'.format(peer)
        command += ' {0} "{1}"'.format(attr, value)
        try:
            out, err = self.execute(command, suppressoutput, suppresserrors)
            if len(err) > 0:
                logger.error('Error executing WireGuard set command: {0}'.format(err))
            else:
                return out
        except Exception as e:
            logger.error('Exception when execution WireGuard set command failed: [{0}]'.format(e))
        return None

    def execute_wg_show(self, suppressoutput=True, suppresserrors=False):
        '''Return the output of "wg show <if> all"'''
        try:
            out, err = self.execute('wg show {0} dump'.format(self.interface), suppressoutput, suppresserrors)
            if len(err) > 0:
                logger.error('Error executing WireGuard command: {0}'.format(err))
            else:
                return out
        except FileNotFoundError:
            logger.error('WireGuard command not found in search path. Is WireGuard installed on this system?')
        return None
        
    def parse_wg_output(self, output):
        '''Parses the given output of the WireGuard command and stores it'''
        if output is None:
            return
        previous_interface = ''
        for line in output.splitlines():
            items = [item for item in line.split('\t')]
            if self.interface == 'all':
                interface = items.pop(0)
            else:
                interface = self.interface
            if interface != previous_interface: # new section with interface information?
                previous_interface = interface
                items = dict(zip(['private-key', 'public-key', 'listen-port', 'fwmark'], items))
                if items['fwmark'] == 'off':
                    items['fwmark'] = None
                items['listen-port'] = int(items['listen-port'])
                if items['private-key'] == '(none)':
                    items['private-key'] = None
                if items['public-key'] == '(none)':
                    items['public-key'] = None
                self.data[interface] = {**self.data.get(interface, dict()), **items} # merge dictionaries
                self.data[interface]['peers'] = self.data[interface].get('peers', dict()) # make sure that peers dictionary exists
            else: # peer
                peer = items.pop(0)
                peerdata = dict(zip(['preshared-key', 'endpoint', 'allowed-ips', 'latest-handshake', 'transfer-rx', 'transfer-tx', 'persistent-keepalive'], items))
                peerdata['allowed-ips'] = peerdata['allowed-ips'].split(',')
                if peerdata['preshared-key'] == '(none)':
                    peerdata['preshared-key'] = None
                if peerdata['persistent-keepalive'] == 'off':
                    peerdata['persistent-keepalive'] = None
                peerdata = { k: int(v) if (k in ['latest-handshake', 'transfer-rx', 'transfer-tx', 'persistent-keepalive']) and (v is not None) else v
                             for k, v in peerdata.items() } # numeric strings to integer
                peerdata['latest-handshake-seconds'], peerdata['handshake-status'] = self.check_handshake(peerdata['latest-handshake'], peerdata['persistent-keepalive'])    
                peerdata['timestamp'] = time.time()
                self.data[interface]['peers'][peer] = {**self.data[interface]['peers'].get(peer, dict()), **peerdata} # merge dictionaries

    def retrieve_wireguard_data(self, data=None):
        '''Sets the local data based on output of WireGuard command to be executed'''
        output = self.execute_wg_show()
        if data is None:
            self.clear_data()
        else:
            self.data = data
        self.parse_wg_output(output)

    @property
    def wgdata():
        return self.data

    def get_interfaces(self,):
        '''Returns the list of interfaces (as an iterator)'''
        return self.data.keys()
        
    def get_interfacedata(self, interface):
        '''Returns the data of the given interface as dictionary'''
        return self.data.get(interface)
        
    def get_peerdata(self, interface):
        '''Returns the data of the peers of the given interface as dictionary'''
        return self.data.get(interface, dict()).get('peers')


if __name__ == '__main__':
    wg = WireguardCommand('all')
    wg.retrieve_wireguard_data()
    pprint.pprint(wg.get_interfaces())
    pprint.pprint(wg.get_interfacedata('wg_1'))
    pprint.pprint(wg.get_peerdata('wg_1'))



# -*- coding: utf-8 -*-

"""Class that maintains the data (attributes and values of interfaces and peers) and makes it accessible"""

import collections
import enum
import logging
import os
import pprint

from . import wg_config
from . import wg_command


logger = logging.getLogger(__name__);


class DataKeeper():
    '''Class for maintaing and accessing the data (attributes and values of interfaces and peers)'''

    def __init__(self, config):
        '''Constructor'''
        self.cfg = config
        self.configfile = config['general']['configfile']        
        self.initialize()

    def initialize(self):
        '''Reads the config and initializes the data structures'''
        # WireGuard command for status information
        self.wgcmd = wg_command.WireguardCommand()
        self.wgcmd.retrieve_wireguard_data()
        self.data = self.wgcmd.data
        # WireGuard config files
        self.wgcfg = wg_config.WireguardConfig()
        self.wgcfg.retrieve_wireguard_config(self.get_interfaces(), self.data, 'config_')
        # Merge config file content into data tree
        self.cfg.read(self.configfile)
        for section in self.cfg.sections():
            if section == 'general':
                for k, v in self.cfg[section].items():
                    self.data[k] = v
            elif section.startswith('interface:'):
                for k, v in self.cfg[section].items():
                    self.data[section][k[10:]] = v
        
    def get(self, interface, peer, attr, default=None):
        '''Gets data matching the specified interface, peer, and attribute (None is allowed for each parameter to get all)'''
        if interface is None:
            if peer is not None:
                raise ValueError('Peer provided but not interface')
            value = self.data
        else:
            value = self.data.get(interface, dict())
            if (attr is not None) or (peer is not None):
                value = value.get(attr if peer is None else 'peers', dict())
                if peer is not None:
                    value = value.get(peer, dict())
                    if attr is not None:
                        value = value.get(attr, default)
        return value    

    def set(self, interface, peer, attr, value):
        '''Sets the value of an attribute matching the specified interface and peer (None is allowed for these)'''
        if (attr is None) or (value is None):
                raise ValueError('Attribute and value need to be provided')
        if interface is None:
            if peer is not None:
                raise ValueError('Peer provided but not interface')
            self.data[attr] = value
        else:
            if peer is None:
                self.data[interface][attr] = value
            else:
                self.data[interface]['peers'][peer][attr] = value
            
    def decrement(self, interface, peer, attr):
        '''Decreases the value of the specified attribute'''
        self.set(interface, peer, attr, self.get(interface, peer, attr, 0) - 1)
            
    def get_interfaces(self):
        '''Returns the list of interfaces (as an iterator)'''
        return self.data.keys()

    def peeriterator(self):
        '''Returns the peer data (as a generator)'''
        for interface, data in self.data.items():
            if isinstance(data, collections.abc.Mapping): # is it a dictionary?
                interfacedata = { k: v for k, v in data.items() if k != 'peers' }
                for peer, peerdata in data.get('peers', dict()).items():
                    yield interface, interfacedata, peer, peerdata

    def update_status(self):
        '''Updates the WireGuard status data'''
        self.wgcmd.retrieve_wireguard_data(self.data)

    def set_endpoint(self, interface, peer, endpoint):
        '''Updates the endpoint of the specified WireGuard peer'''
        return self.wgcmd.execute_wg_set(interface, peer, 'endpoint', endpoint)


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-

import contextlib
import os
import stat
import sys
import tempfile


@contextlib.contextmanager
def open_for_atomic_write(filename, text=True, uid=None, gid=None, perm=None, tmp_suffix='.bak', tmp_prefix='tmp_', tmp_keep=True):
    '''Context manager for overwriting a file atomically.'''
    # Create temporary file
    path = os.path.dirname(filename)
    fd, filetmp = tempfile.mkstemp(dir=path, text=text, suffix=tmp_suffix, prefix=tmp_prefix)
    try:
        # Open remporary file for writing
        with os.fdopen(fd, 'w' if text else 'wb') as f:
            yield f
        # Rename to target
        os.replace(filetmp, filename) # atomic on POSIX systems and Windows for Python 3.3+
        filetmp = None
        # Set owner and permissions
        if any(x is None for x in (uid, gid, perm)):
            try:
                info = os.stat(filename)
                if uid is None:
                    uid = info.st_uid
                if gid is None:
                    gid = info.st_gid
                if perm is None:
                    perm = stat.S_IMODE(info.st_mode)
            except FileNotFoundError:
                pass
        if uid is None:
            uid = -1
        if gid is None:
            gid = -1
        os.chown(filename, uid, gid)
        if perm is not None:
            os.chmod(filename, perm)
    finally: # Silently try to delete the temporary file if needed
        if (filetmp is not None) and not tmp_keep: 
            try:
                os.unlink(filetmp)
            except:
                pass
